#!/usr/bin/env python
# -*- coding: UTF-8 -*-

##
## © Copyright 2021- IBM Inc. All rights reserved
# SPDX-License-Identifier: MIT
##


app = 'elmoslcquery'
description = 'Commandline OSLC query for ELM'
version = '0.6.1'
license = 'MIT'
author_name = 'Ian Barnard'
author_mail = 'ian.barnard@uk.ibm.com'
author = '%s <%s>' % (author_name, author_mail)
url = 'https://github.com/IBM/ELM-Python-Client'
tarball = '%s/archive/%s.tar.gz' % (url, version)
